##ºº startR - Retrieval and processing of multidimensional datasets

Welcome to startR! startR is an R package developed at the Barcelona Supercomputing Center which implements the MapReduce paradigm (a.k.a. domain decomposition) on HPCs in a way transparent to the users and specially tailored for complex multidimensional datasets. startR is an open source project which is open to external collaboration and funding, and will continuously evolve to support as many data set formats as possible while maximizing its efficiency.

### How it works
The startR framework can be summarized in the following steps:
1. [Declaration of data sources](inst/doc/practical_guide.md#1-declaration-of-data-sources)
2. [Declaration of the workflow](inst/doc/practical_guide.md#2-declaration-of-the-workflow)
3. [Declaration of the HPC platform and execution](inst/doc/practical_guide.md#3-declaration-of-the-hpc-platform-and-execution)
4. [Monitoring the execution](inst/doc/practical_guide.md#4-monitoring-the-execution)
5. [Profiling of the execution](inst/doc/practical_guide.md#5-profiling-of-the-execution)

To let the new users have a preliminary understanding of the script, an example for each step is shown below. It shows a simple addition and average operation performed on BSC's CTE-Power HPC (i.e., Power 9) over a multi-dimensional climate data set, which is stored in the BSC-ES storage infrastructure. The purpose of this example is simply to illustrate how the users can expect from using startR. You can find real use cases in [practical_guide.md](inst/doc/practical_guide.md).

#### 1. Declaration of data sources
As mentioned above, the first step is to declare the data sources. With `retrieve = FALSE`, it creates a pointer indicating the desired dataset.

```r
library(startR)

# A path pattern is built with self-defined wildcards
repos <- '/esarchive/exp/ecmwf/system5_m1/6hourly/$var$/$var$_$sdate$.nc'

# A Start() call is built with the indices to operate
data <- Start(dat = repos,
              var = 'tas',
              sdate = '20180101',
              ensemble = 'all',
              time = 'all',
              latitude = indices(1:40),
              longitude = indices(1:40),
              retrieve = FALSE)
```

#### 2. Declaration of the workflow
The second step is to define the function and build up the workflow using `Step()` and `AddStep()`.

```r
# The function to be applied is defined
# It only operates on the essential 'target' dimensions
fun <- function(x) {
  # Expected inputs:
  #   x: array with dimensions ('ensemble', 'time')
  apply(x + 1, 2, mean)
  # Outputs:
  #   single array with dimensions ('time')
}

# A startR Step is defined, specifying its expected input and
# output dimensions
step <- Step(fun,
             target_dims = c('ensemble', 'time'),
             output_dims = c('time'))

# The workflow of operations is cast before execution
wf <- AddStep(data, step)
```

#### 3. Declaration of the HPC platform and execution
After data sources and workflow are prepared, we are ready to dispatch jobs to the wanted machines, which is Power 9 in this case. The computation configuration of Power 9 is defined in the `cluster` list.

```r
res <- Compute(wf,
               chunks = list(latitude = 2,
                             longitude = 2),
               threads_load = 1,
               threads_compute = 2,
               cluster = list(queue_host = 'p9login1.bsc.es',  #user-defined
                              queue_type = 'slurm',
                              temp_dir = '/gpfs/scratch/bsc32/bsc32473/startR_tests/', #user-defined
                              lib_dir = '/gpfs/projects/bsc32/share/R_libs/3.5/',
                              r_module = 'R/3.5.0-foss-2018b',
                              cores_per_job = 2,
                              job_wallclock = '00:10:00',
                              max_jobs = 4,
                              extra_queue_params = list('#SBATCH --qos=bsc_es'),
                              bidirectional = FALSE,
                              polling_period = 10
                             ),
               ecflow_suite_dir = '/home/Earth/nmanuben/test_startR/',  #user-defined
               wait = TRUE)
```

#### 4. Monitoring the execution

During the execution of the workflow, which is orchestrated by EC-Flow and a job scheduler (either Slurm, LSF or PBS), the status can be monitored using the EC-Flow graphical user interface (see figure below). 

<img src="inst/doc/ecflow_monitoring.png" width="600" />

#### 5. Profiling of the execution

Additionally, profiling measurements of the execution are provided together with the output data. The profiling will be printed on the R terminal when the execution ends, or you can visualize it with the `PlotProfiling()` function available in the source code of the startR package.

```r
source('https://earth.bsc.es/gitlab/es/startR/raw/master/inst/PlotProfiling.R')
PlotProfiling(attr(res, 'startR_compute_profiling'))
```

You can click on the image to expand it.

<img src="inst/doc/compute_profiling.png" width="800" />



### Installation
Now you have a general picture of how startR works. But before starting using startR, you need to check if the required environments are installed. The most relevant system dependencies are listed below:
- netCDF-4
- R, with the startR packages and its dependent packages (e.g., bigmemory, easyNCDF)
- For distributed computation:
  - UNIX-based HPC (cluster of multi-processor nodes)
  - a job scheduler (Slurm, PBS or LSF)
  - EC-Flow >= 4.9.0

### Other useful documentation
If you are a new startR users, you are recommended to go through the following link one by one to get familiar with startR.

1. See [**Deployment**](inst/doc/deployment.md) for details on the setup if outside of BSC.  
2. Follow [**Practical Guide**](inst/doc/practical_guide.md) with use cases step by step.
3. See [**Start**](inst/doc/start.md) to learn more applications of Start().
4. Find [**FAQs**](https://earth.bsc.es/gitlab/es/startR/wikis/FAQ) and [**Examples**](https://earth.bsc.es/gitlab/es/startR/wikis/Example) in GitLab wiki page.

